from typing import Any, Callable, Optional, Type

import attr
from marshmallow import Schema

from rho_ml.utils import Version


class ValidationFailedError(Exception):
    pass


# todo: consider alternate name for Predictor
@attr.s(auto_attribs=True)
class RhoModel(object):
    """ The most basic wrapper for an ML or other model.  Specific models should
    be a subclass of Predictor, and implement whatever appropriate subset of
    abstract methods defined here """
    cached_version: Version = None

    @property
    def name(self) -> str:
        return str(self.__class__.__name__)
    
    @property
    def version(self) -> Version:
        if self.cached_version:
            return self.cached_version
        else:
            raise NotImplementedError("No model version set!")

    @version.setter
    def version(self, value: Version):
        self.cached_version = value
    
    @property
    def training_data_schema(self) -> Optional[Type[Schema]]:
        return None
    
    def validate_training_input(self, data: Any):
        """ Throw a ValidationFailedError if the data is not valid training data """
        raise NotImplementedError
    
    def validate_training_output(self, data: Any):
        """ Throw a ValidationFailedError if the data is not valid output from the train()
        method. """
        raise NotImplementedError
    
    def validate_prediction_input(self, data: Any):
        """ Throw a ValidationFailedError if the data is not valid prediction input data """
        raise NotImplementedError
    
    def validate_prediction_output(self, data: Any):
        """ Throw a ValidationFailedError if the data is not valid prediction output
        data """
        raise NotImplementedError
    
    def train_logic(self, training_data: Any, *args, **kwargs) -> Any:
        """ This method should be overridden with the appropriate logic to
        take training data, evaluation data, run training, and return
        relevant data (e.g. training and validation metrics for each epoch)."""
        raise NotImplementedError
    
    def train(self, training_data: Any, run_validation: bool,
              *args, **kwargs) -> Any:
        if run_validation:
            result = get_validated_result(
                model_method=self.train_logic,
                input_data=training_data,
                input_validator=self.validate_training_input,
                output_validator=self.validate_training_output,
                *args, **kwargs)
        else:
            result = self.train_logic(
                training_data=training_data, *args, **kwargs)
        return result
    
    def predict_logic(self, prediction_data: Any, *args, **kwargs) -> Any:
        """ This method should take a dict containing items with appropriate
        types for the model, and generate outputs with appropriate types."""
        raise NotImplementedError
    
    def predict(self, prediction_data: Any, run_validation: bool,
                *args, **kwargs) -> Any:
        if run_validation:
            result = get_validated_result(
                model_method=self.predict_logic,
                input_data=prediction_data,
                input_validator=self.validate_prediction_input,
                output_validator=self.validate_prediction_output,
                *args, **kwargs)
        else:
            result = self.predict_logic(
                prediction_data=prediction_data, *args, **kwargs)
        return result
    
    def serialize(self) -> bytes:
        """ Serialize the predictor object to a byte string"""
        raise NotImplementedError
    
    @classmethod
    def deserialize(cls, serialized: bytes):
        """ Instantiate a Predictor object from a serialized byte string """
        raise NotImplementedError
    
    def save_to_disk(self, path_to_output: str):
        """ Logic to save a single Predictor object to disk."""
        serialized = self.serialize()
        with open(path_to_output, 'wb') as f:
            f.write(serialized)
    
    @classmethod
    def load_from_disk(cls, path_to_file: str):
        """ Logic to load the Predictor subclass from disk. """
        with open(path_to_file, 'rb') as f:
            loaded = cls.deserialize(f.read())
        return loaded


def get_validated_result(model_method: Callable,
                         input_data: Any,
                         input_validator: Callable[[Any], None],
                         output_validator: Callable[[Any], None],
                         *args, **kwargs) -> Any:
    input_validator(input_data)
    output_data = model_method(input_data, *args, **kwargs)
    output_validator(output_data)
    return output_data



from typing import Optional

import attr
import re


def convert_empty_tag(tag: Optional[str]) -> Optional[str]:
    if not tag:
        result = None
    else:
        result = tag
    return result


@attr.s(auto_attribs=True, frozen=True)
class Version(object):
    major: int
    minor: int
    patch: int
    tag: Optional[str] = attr.ib(converter=convert_empty_tag, default=None)

    def __str__(self):
        return self.to_string()

    @classmethod
    def from_string(cls, val: str):
        """ Create a Version object from a string of the form
        'major.minor.patch-tag', where the tag is optional.  If it is present,
        it must be separated from the rest of the string by a hyphen for this
        method to find it."""
        m = re.match(r'(?P<major>[0-9]+)\.(?P<minor>[0-9]+)\.(?P<patch>[0-9]+)'
                     r'(-(?P<tag>.+))?',
                     val)
        return cls(major=int(m.group('major')),
                   minor=int(m.group('minor')),
                   patch=int(m.group('patch')),
                   tag=m.group('tag'))

    def to_string(self) -> str:
        """ Create a version string of the form "major.min.patch-tag, where the
        hyphen and tag are only present if `self.tag != None` """
        version_str = "{0}.{1}.{2}".format(self.major, self.minor, self.patch)
        if self.tag:
            version_str += '-' + self.tag
        return version_str



import base64
import importlib
import json
import pickle
from typing import Optional

import attr

from rho_ml.rho_model import RhoModel

# TODO: docstrings for *all* methods


@attr.s(auto_attribs=True, frozen=True)
class StoredModel(object):
    """ This stores the bytes of a serialized model, along with the module and
    qualified class name needed to correctly deserialize the model later.

    If the base_class is provided, that is the class which will be imported,
    and then wrapped with class_name as a subclass which differs from base_class
    only in name.  This is useful for the case where the same underlying model
    is used for multiple separate applications. """
    module_name: str
    class_name: str
    model_bytes: bytes = attr.ib(repr=False)
    base_class: Optional[str] = None

    @classmethod
    def from_model(cls,
                   model: RhoModel,
                   base_class_name: Optional[str] = None):
        return cls(module_name=model.__module__,
                   class_name=model.__class__.__qualname__,
                   model_bytes=model.serialize(),
                   base_class=base_class_name)

    def load_model(self) -> RhoModel:
        """ Use the instantiated StoredPredictor to load the Predictor using
        the appropriate class. """
        predictor_module = importlib.import_module(self.module_name)
        if self.base_class:
            base_cls = getattr(predictor_module, self.base_class)
            predictor_cls = type(self.class_name, (base_cls,), {})
        else:
            predictor_cls = getattr(predictor_module, self.class_name)
        predictor = predictor_cls.deserialize(self.model_bytes)
        predictor.__module__ = self.module_name
        predictor.__class__.__qualname__ = self.class_name
        return predictor

    def to_pickle(self) -> bytes:
        return pickle.dumps(self, protocol=4)

    @classmethod
    def from_pickle(cls, s: bytes) -> 'StoredModel':
        return pickle.loads(s)

    def to_json(self) -> str:
        """ JSON serialize the StoredPredictor so it can be saved to Redis, S3,
        etc. This uses base64 encoding on the predictor bytes, so it is often
        not going to be the most efficient way to store a given model. """
        # todo: deprecate me
        storage_dict = attr.asdict(self)
        storage_dict['model_bytes'] = base64.encodebytes(
            storage_dict['model_bytes']).decode('ascii')
        return json.dumps(storage_dict)

    @classmethod
    def from_json(cls, storage_json: str):
        # todo: deprecate me
        # add logging warning this is being deprecated
        storage_dict = json.loads(storage_json)
        encoded_model = storage_dict['model_bytes'].encode('ascii')
        storage_dict['model_bytes'] = base64.decodebytes(encoded_model)
        return cls(**storage_dict)
